#include "diagnosis_wizard.h"
#include "prime_factor.h"

#include <utility>
#include <algorithm>

#include "QMessageBox"


using std::pair;
using std::vector;
using std::make_pair;
using std::sort;
using std::unique;
using std::min;

DiagnosisWizard::DiagnosisWizard(QWidget *parent) : Q3Wizard(parent), Ui::DiagnosisWizard()
{
    setupUi(this);
    setNextEnabled(OverlapPage, false);
}


void DiagnosisWizard::setData(std::vector<PartitionTable>* pts, std::vector<QString>* devs)
{
    mPartitionTables = pts;
    mDeviceArray = devs;
    checkOverlap();
}



static bool operator<(const pair<unsigned long int, unsigned long int>& lhs, const pair<unsigned long int, unsigned long int>& rhs)
{
    return lhs.first < rhs.first;
}



static bool isBlank(const PTEntry& pte)
{
    return !(pte.type && pte.lbaStart && pte.lbaLength);
}



void DiagnosisWizard::checkOverlap()
{
    bool allgood = true;
    vector<pair<unsigned long int, unsigned long int> > table;
    for (size_t i = 0; i < mPartitionTables->size(); ++i)
    {
	table.clear();
	for (int j = 0; j < 4; ++j)
	{
            if (!isBlank((*mPartitionTables)[i].mTable[j]))
            {
                table.push_back(make_pair((*mPartitionTables)[i].mTable[j].lbaStart,
                                          (*mPartitionTables)[i].mTable[j].lbaLength));
            } // end if
	} // end for
	sort(table.begin(), table.end());
	
	bool thisgood = true;
	for (int j = 1; j < table.size(); ++j)
	{
	    // check that partition j-1 does not overlap partition j
	    if (table[j-1].first + table[j-1].second > table[j].first)
	    {
		thisgood = false;
		allgood = false;
		break;
	    }
	}
	
	new Q3ListViewItem(overlapListView, (*mDeviceArray)[i], thisgood ? "Good" : "Bad");
    }
    setNextEnabled(OverlapPage, allgood);
}



void DiagnosisWizard::pageChange(const QString& page)
{
    if (page == title(DuplicatePage))
    {
	checkDuplicates();
    }
    else if (page == title(GapPage))
    {
	checkGaps();
    }
    else if (page == title(OrderPage))
    {
        fillOrder();
    }
    else if (page == title(CHSLBAPage))
    {
        checkCHS();
    }
}



static bool eq_notzero(unsigned char c1, unsigned char c2)
{
    return c1 == c2 && c1 != 0;
}



void DiagnosisWizard::checkDuplicates()
{
    bool allgood = true;
    vector<unsigned char> types;
    
    duplicateListView->clear();

    for (size_t i = 0; i < mPartitionTables->size(); ++i)
    {
	types.clear();
	for (int j = 0; j < 4; ++j)
	{
	    types.push_back((*mPartitionTables)[i].mTable[j].type);
	} // end for
	sort(types.begin(), types.end());
	types.erase(unique(types.begin(), types.end(), eq_notzero), types.end());
	
	bool thisgood = true;
	if (4 != types.size())
	{
            thisgood = false;
            allgood = false;
        }

	new Q3ListViewItem(duplicateListView, (*mDeviceArray)[i], thisgood ? "Good" : "Bad");
    }
    setNextEnabled(DuplicatePage, allgood);
}



void DiagnosisWizard::checkGaps()
{
    vector<pair<unsigned long int, unsigned long int> > table;
    gapListView->clear();
    for (size_t i = 0; i < mPartitionTables->size(); ++i)
    {
	table.clear();
	for (int j = 0; j < 4; ++j)
	{
            if (!isBlank((*mPartitionTables)[i].mTable[j]))
            {
                table.push_back(make_pair((*mPartitionTables)[i].mTable[j].lbaStart,
                                          (*mPartitionTables)[i].mTable[j].lbaLength));
            }
	} // end for
	sort(table.begin(), table.end());

	bool gaps = false;

	for (int j = 1; j < table.size(); ++j)
	{
	    // check that partition j-1 starts just after
	    if (table[j-1].first + table[j-1].second != table[j].first)
	    {
		gaps = true;
	    }
	}

	new Q3ListViewItem(gapListView, (*mDeviceArray)[i], gaps ? "Gaps" : "No Gaps");
    }
}



void DiagnosisWizard::fillOrder()
{
    vector<unsigned long int> starts;
    orderListView->clear();
    for (size_t i = 0; i < mPartitionTables->size(); ++i)
    {
	starts.clear();
	for (int j = 0; j < 4; ++j)
	{
            if (!isBlank((*mPartitionTables)[i].mTable[j]))
            {
                starts.push_back((*mPartitionTables)[i].mTable[j].lbaStart);
            } // end if
	} // end for
	
	bool inorder = true;
	for (int j = 1; j < starts.size(); ++j)
	{
	    if (starts[j-1] > starts[j])
	    {
		inorder = false;
		break;
	    }
	}
	
	new Q3ListViewItem(orderListView, (*mDeviceArray)[i], inorder ? "In Order" : "Out of Order");
    }
    orderListView->setSelected(orderListView->firstChild(), true);
}


void DiagnosisWizard::setCHS()
{
    // set the values
    checkOrSetCHS(false);
}


static bool operator<(const PTEntry& pte1, const PTEntry& pte2)
{
    return pte1.lbaStart < pte2.lbaStart;
}


void DiagnosisWizard::repairOrder()
{
    Q3ListViewItem* i = orderListView->selectedItem();

    if (i->text(1) == "In Order")
    {
        QMessageBox::warning(this, "Warning", "Nothing to do");
        return;
    }

    int index = 0;
    for (vector<QString>::const_iterator iter = mDeviceArray->begin();
         iter != mDeviceArray->end();
         ++iter, ++index)
    {
        if (*iter == i->text(0))
        {
            break;
        } // end if
    } // end for

    sort(&((*mPartitionTables)[index].mTable[0]), &((*mPartitionTables)[index].mTable[3]));

}


void DiagnosisWizard::checkCHS()
{
    CHSLBAListView->clear();
    checkOrSetCHS(true);
}



#include <iostream>
using namespace std;
void DiagnosisWizard::checkOrSetCHS(bool check)
{
    // values stored in partition tables
    unsigned long int C, H, S;
    // values obtained from factorizing LBA
    unsigned long int cyl, head, sec;
    const unsigned long int CHSmax = 1024ul*256ul*63ul;
    bool allGood = true;
    QString changes;
    for (size_t i = 0; i < mPartitionTables->size(); ++i)
    {
        bool currentPartitionGood = true;
        cout << (*mDeviceArray)[i].toLocal8Bit().constData() << endl;
        for (int j = 0; j < 4; ++j)
	{
            if (!isBlank((*mPartitionTables)[i].mTable[j]))
            {
                PTEntry& pte = (*mPartitionTables)[i].mTable[j];

                // check end
                C = pte.cylinderSectorEnd.extractedData.cylinder;
                H = pte.headEnd;
                S = pte.cylinderSectorEnd.extractedData.sector;
                getCHS(cyl, head, sec, pte.lbaStart+pte.lbaLength);
                // what I think the values should be
                unsigned long int CEndGuess, HEndGuess, SEndGuess;
                if (pte.lbaStart+pte.lbaLength >= CHSmax)
                {
                    CEndGuess = 1023ul;
                    HEndGuess = 254ul;
                    SEndGuess = 63ul;
                }
                else
                {
                    CEndGuess = min(cyl-1, 1023ul);
                    HEndGuess = head-1;
                    SEndGuess = S;
                }
                cout << "end LBA " << pte.lbaStart+pte.lbaLength << endl;
                cout << "C " << C << " H " << H << " S " << S;
                cout << " cyl " << cyl << " head " << head << " sec " << sec << endl;
                cout << "my guesses CEndGuess " << CEndGuess
                     << " HEndGuess " << HEndGuess
                     << " SEndGuess " << SEndGuess << endl;
                if (C != CEndGuess ||
                    H != HEndGuess ||
                    S != SEndGuess)
                {
                    currentPartitionGood = false;
                    allGood = false;
                    if (!check)
                    {
                        changes += QString("%1: End C,H,S %2,%3,%4 changed to %5,%6,%7\n")
                            .arg((*mDeviceArray)[i])
                            .arg(pte.cylinderSectorEnd.extractedData.cylinder)
                            .arg(pte.headEnd)
                            .arg(pte.cylinderSectorEnd.extractedData.sector)
                            .arg(CEndGuess)
                            .arg(HEndGuess)
                            .arg(SEndGuess);
                            
                        pte.cylinderSectorEnd.extractedData.cylinder = CEndGuess;
                        pte.headEnd = HEndGuess;
                        pte.cylinderSectorEnd.extractedData.sector = SEndGuess;
                    }
                }
                // check start
                C = pte.cylinderSectorStart.extractedData.cylinder;
                H = pte.headStart;
                S = pte.cylinderSectorStart.extractedData.sector;
                getCHS(cyl, head, sec, pte.lbaStart);
                unsigned long int CStartGuess, HStartGuess, SStartGuess;
                if (pte.lbaStart >= CHSmax)
                {
                    CStartGuess = 1023ul;
                    HStartGuess = 254ul;
                    SStartGuess = 63ul;
                }
                else
                {
                    if (cyl == 1 && head == 1 && sec == SEndGuess)
                    {
                        // special case, 1st partition starting
                        // just after MBR
                        CStartGuess = 0;
                        HStartGuess = 1;
                        SStartGuess = 1;
                    }
                    else
                    {
                        CStartGuess = cyl;
                        HStartGuess = head/HEndGuess - 1ul;;
                        SStartGuess = sec/SEndGuess;
                    }
                }
                cout << "start LBA " << pte.lbaStart << endl;
                cout << "C " << C << " H " << H << " S " << S;
                cout << " cyl " << cyl << " head " << head << " sec " << sec << endl;
                cout << "my guesses CStartGuess " << CStartGuess
                     << " HStartGuess " << HStartGuess
                     << " SStartGuess " << SStartGuess << endl;

                if (C != CStartGuess ||
                    H != HStartGuess ||
                    S != SStartGuess)
                {
                    currentPartitionGood = false;
                    allGood = false;
                    if (!check)
                    {
                        changes += QString("%1: Start C,H,S %2,%3,%4 changed to %5,%6,%7\n")
                            .arg((*mDeviceArray)[i])
                            .arg(pte.cylinderSectorStart.extractedData.cylinder)
                            .arg(pte.headStart)
                            .arg(pte.cylinderSectorStart.extractedData.sector)
                            .arg(CStartGuess)
                            .arg(HStartGuess)
                            .arg(SStartGuess);

                        pte.cylinderSectorStart.extractedData.cylinder = CStartGuess;
                        pte.headStart = HStartGuess;
                        pte.cylinderSectorStart.extractedData.sector = SStartGuess;
                    }
                }
            } // end if
        } // end for
        if (check)
        {
            new Q3ListViewItem(CHSLBAListView, (*mDeviceArray)[i], currentPartitionGood ? "Match" : "Mismatch");
        } // end if
    } // end for
    if (check)
    {
        CHSPushButton->setEnabled(!allGood);
        setFinishEnabled(CHSLBAPage, allGood);
    }
    else
    {
        QMessageBox::information(this, "Changes Applied", changes);
        setFinishEnabled(CHSLBAPage, true);
    } // end if
}
